@@ -31,8 +31,6 @@
 
 #include <sys/types.h>
 #include <sys/socket.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
 
 #include <ctype.h>
 #include <stdio.h>
@@ -52,13 +50,14 @@ extern int expand_special_asn;
 static int
 usage(int ecode)
 {
-	printf("\nUsage: bgpq4 [-h host[:port]] [-S sources] [-E|G <num>"
+	printf("\nUsage: bgpq4 [-h host[:port]] [-S sources] [-E|G|H <num>"
 	    "|f <num>|t] [-46ABbdJjKNnwXz] [-R len] <OBJECTS> ... "
 	    "[EXCEPT <OBJECTS> ...]\n");
 	printf("\nVendor targets:\n");
 	printf(" no option : Cisco IOS Classic (default)\n");
 	printf(" -X        : Cisco IOS XR\n");
 	printf(" -U        : Huawei\n");
+	printf(" -u        : Huawei XPL\n");
 	printf(" -j        : JSON\n");
 	printf(" -J        : Juniper Junos\n");
 	printf(" -K        : MikroTik RouterOS\n");
@@ -88,6 +87,7 @@ usage(int ecode)
 	    "(OpenBGPD)\n");
 	printf(" -f number : generate input as-path access-list\n");
 	printf(" -G number : generate output as-path access-list\n");
+	printf(" -H number : generate origin as-lists (JunOS only)\n");
 	printf(" -M match  : extra match conditions for JunOS route-filters\n");
 	printf(" -l name   : use specified name for generated access/prefix/.."
 		" list\n");
@@ -97,7 +97,7 @@ usage(int ecode)
 	printf(" -t        : generate as-sets for OpenBGPD (OpenBGPD 6.4+), BIRD "
 		"and JSON formats\n");
 	printf(" -z        : generate route-filter-list (Junos only)\n");
-	printf(" -W len    : specify max-entries on as-path line (use 0 for "
+	printf(" -W len    : specify max-entries on as-path/as-list line (use 0 for "
 		"infinity)\n");
 
 	printf("\nUtility operations:\n");
@@ -134,8 +134,8 @@ vendor_exclusive(void)
 {
 	fprintf(stderr, "-b (BIRD), -B (OpenBGPD), -F (formatted), -J (Junos),"
 	    " -j (JSON), -N (Nokia SR OS Classic), -n (Nokia SR OS MD-CLI),"
-	    " -U (Huawei), -e (Arista) and -X (IOS XR) options are mutually"
-	    " exclusive\n");
+	    " -U (Huawei), -u (Huawei XPL), -e (Arista) and -X (IOS XR) options "
+	    " are mutually exclusive\n");
 	exit(1);
 }
 
@@ -197,149 +197,149 @@ main(int argc, char* argv[])
 		expander.sources=getenv("IRRD_SOURCES");
 
 	while ((c = getopt(argc, argv,
-	    "346a:AbBdDEeF:S:jJKf:l:L:m:M:NnW:pr:R:G:tTh:UwXsvz"))
-	    !=EOF) {
+	    "46a:AbBdDEeF:S:jJKf:l:L:m:M:NnW:pr:R:G:H:tTh:UuwXsvz")) != EOF) {
 	switch (c) {
-        case '3':
-            /*
-             * No-op, left for backwards compatibility with bgpq3
-             */
-            break;
-		case '4':
-			/* do nothing, expander already configured for IPv4 */
-			if (expander.family == AF_INET6) {
-				sx_report(SX_FATAL, "-4 and -6 are mutually "
-				    "exclusive\n");
-				exit(1);
+	case '4':
+		/* do nothing, expander already configured for IPv4 */
+		if (expander.family == AF_INET6) {
+			sx_report(SX_FATAL, "-4 and -6 are mutually "
+			    "exclusive\n");
+			exit(1);
+		}
+		selectedipv4 = 1;
+		break;
+	case '6':
+		if (selectedipv4) {
+			sx_report(SX_FATAL, "-4 and -6 are mutually "
+			    "exclusive\n");
+			exit(1);
+		}
+		af = AF_INET6;
+		expander.family = AF_INET6;
+		expander.tree->family = AF_INET6;
+		break;
+	case 'a':
+		parseasnumber(&expander, optarg);
+		break;
+	case 'A':
+		if (aggregate)
+			debug_aggregation++;
+		aggregate = 1;
+		break;
+	case 'b':
+		if (expander.vendor)
+			vendor_exclusive();
+		expander.vendor = V_BIRD;
+		break;
+	case 'B':
+		if (expander.vendor)
+			vendor_exclusive();
+		expander.vendor = V_OPENBGPD;
+		break;
+	case 'd':
+		debug_expander++;
+		break;
+	case 'E':
+		if (expander.generation)
+			exclusive();
+		expander.generation = T_EACL;
+		break;
+	case 'e':
+		if (expander.vendor)
+			vendor_exclusive();
+		expander.vendor = V_ARISTA;
+		expander.sequence = 1;
+		break;
+	case 'F':
+		if (expander.vendor)
+			exclusive();
+		expander.vendor = V_FORMAT;
+		expander.format = optarg;
+		break;
+	case 'f':
+		if (expander.generation)
+			exclusive();
+		expander.generation = T_ASPATH;
+		parseasnumber(&expander, optarg);
+		break;
+	case 'G':
+		if (expander.generation)
+			exclusive();
+		expander.generation = T_OASPATH;
+		parseasnumber(&expander, optarg);
+		break;
+	case 'H':
+		if (expander.generation)
+			exclusive();
+		expander.generation = T_ASLIST;
+		parseasnumber(&expander, optarg);
+		break;
+	case 'h':
+		{
+			char *d = strchr(optarg, ':');
+			expander.server = optarg;
+			if (d) {
+				*d = 0;
+				expander.port = d + 1;
 			}
-			selectedipv4=1;
-			break;
-		case '6':
-			if (selectedipv4) {
-				sx_report(SX_FATAL, "-4 and -6 are mutually "
-				    "exclusive\n");
-				exit(1);
-			}
-			af = AF_INET6;
-			expander.family = AF_INET6;
-			expander.tree->family = AF_INET6;
-			break;
-		case 'a':
-			parseasnumber(&expander, optarg);
-			break;
-		case 'A':
-			if (aggregate)
-				debug_aggregation++;
-			aggregate = 1;
-			break;
-		case 'b':
-			if (expander.vendor)
-				vendor_exclusive();
-			expander.vendor = V_BIRD;
-			break;
-		case 'B':
-			if (expander.vendor)
-				vendor_exclusive();
-			expander.vendor = V_OPENBGPD;
-			break;
-		case 'd':
-			debug_expander++;
-			break;
-		case 'E':
-			if (expander.generation)
-				exclusive();
-			expander.generation = T_EACL;
-			break;
-		case 'e':
-			if (expander.vendor)
-				vendor_exclusive();
-			expander.vendor = V_ARISTA;
-			expander.sequence = 1;
-			break;
-		case 'F':
-			if (expander.vendor)
-				exclusive();
-			expander.vendor = V_FORMAT;
-			expander.format = optarg;
-			break;
-		case 'f':
-			if (expander.generation)
-				exclusive();
-			expander.generation = T_ASPATH;
-			parseasnumber(&expander, optarg);
-			break;
-		case 'G':
-			if (expander.generation)
-				exclusive();
-			expander.generation = T_OASPATH;
-			parseasnumber(&expander, optarg);
-			break;
-		case 'h':
-			{
-				char* d = strchr(optarg, ':');
-				expander.server = optarg;
-				if (d) {
-					*d = 0;
-					expander.port = d + 1;
-				}
-			}
-			break;
-		case 'J':
-			if (expander.vendor)
-				vendor_exclusive();
-			expander.vendor = V_JUNIPER;
-			break;
-		case 'j':
-			if (expander.vendor)
-				vendor_exclusive();
-			expander.vendor = V_JSON;
-			break;
-		case 'K':
-			if (expander.vendor)
-				vendor_exclusive();
-			expander.vendor = V_MIKROTIK;
-			break;
-		case 'p':
-			expand_special_asn = 1;
-			break;
-		case 'r':
-			refineLow = strtoul(optarg, NULL, 10);
-			if (!refineLow) {
-				sx_report(SX_FATAL, "Invalid refineLow value:"
-				    " %s\n", optarg);
-				exit(1);
-			}
-			break;
-		case 'R':
-			refine = strtoul(optarg, NULL, 10);
-			if (!refine) {
-				sx_report(SX_FATAL,"Invalid refine length:"
-				    " %s\n", optarg);
-				exit(1);
-			}
-			break;
-		case 'l':
-			expander.name = optarg;
-			break;
-		case 'L':
-			expander.maxdepth = strtol(optarg, NULL, 10);
-			if (expander.maxdepth < 1) {
-				sx_report(SX_FATAL, "Invalid maximum recursion"
-				    " (-L): %s\n", optarg);
-				exit(1);
-			}
-			break;
-		case 'm':
-			maxlen=strtoul(optarg, NULL, 10);
-			if (!maxlen) {
-				sx_report(SX_FATAL, "Invalid maxlen (-m): %s\n",
-				    optarg);
-				exit(1);
-			}
-			break;
-		case 'M':
-			{
-			char *mc, *md;
+		}
+		break;
+	case 'J':
+		if (expander.vendor)
+			vendor_exclusive();
+		expander.vendor = V_JUNIPER;
+		break;
+	case 'j':
+		if (expander.vendor)
+			vendor_exclusive();
+		expander.vendor = V_JSON;
+		break;
+	case 'K':
+		if (expander.vendor)
+			vendor_exclusive();
+		expander.vendor = V_MIKROTIK;
+		break;
+	case 'p':
+		expand_special_asn = 1;
+		break;
+	case 'r':
+		refineLow = strtoul(optarg, NULL, 10);
+		if (!refineLow) {
+			sx_report(SX_FATAL, "Invalid refineLow value:"
+			    " %s\n", optarg);
+			exit(1);
+		}
+		break;
+	case 'R':
+		refine = strtoul(optarg, NULL, 10);
+		if (!refine) {
+			sx_report(SX_FATAL,"Invalid refine length:"
+			    " %s\n", optarg);
+			exit(1);
+		}
+		break;
+	case 'l':
+		expander.name = optarg;
+		break;
+	case 'L':
+		expander.maxdepth = strtol(optarg, NULL, 10);
+		if (expander.maxdepth < 1) {
+			sx_report(SX_FATAL, "Invalid maximum recursion"
+			    " (-L): %s\n", optarg);
+			exit(1);
+		}
+		break;
+	case 'm':
+		maxlen=strtoul(optarg, NULL, 10);
+		if (!maxlen) {
+			sx_report(SX_FATAL, "Invalid maxlen (-m): %s\n",
+			    optarg);
+			exit(1);
+		}
+		break;
+	case 'M':
+		{
+			char	*mc, *md;
 			expander.match = strdup(optarg);
 			mc = md = expander.match;
 			while (*mc) {
@@ -377,64 +377,68 @@ main(int argc, char* argv[])
 				}
 			}
 			*md = 0;
-			}
-			break;
-		case 'N':
-			if (expander.vendor)
-				vendor_exclusive();
-			expander.vendor = V_NOKIA;
-			break;
-		case 'n':
-			if (expander.vendor)
-				vendor_exclusive();
-			expander.vendor = V_NOKIA_MD;
-			break;
-		case 't':
-			if (expander.generation)
-				exclusive();
-			expander.generation = T_ASSET;
-			break;
-		case 'T':
-			pipelining = 0;
-			break;
-		case 's':
-			expander.sequence = 1;
-			break;
-		case 'S':
-			expander.sources = optarg;
-			break;
-		case 'U':
-			if (expander.vendor)
-				exclusive();
-			expander.vendor = V_HUAWEI;
-			break;
-		case 'W':
-			expander.aswidth = atoi(optarg);
-			if (expander.aswidth < 0) {
-				sx_report(SX_FATAL,"Invalid as-width: %s\n",
-				    optarg);
-				exit(1);
-			}
-			widthSet = 1;
-			break;
-		case 'w':
-			expander.validate_asns = 1;
-			break;
-		case 'X':
-			if (expander.vendor)
-				vendor_exclusive();
-			expander.vendor = V_CISCO_XR;
-			break;
-		case 'v':
-			version();
-			break;
-		case 'z':
-			if (expander.generation)
-				exclusive();
-			expander.generation = T_ROUTE_FILTER_LIST;
-			break;
-		default:
-			usage(1);
+		}
+		break;
+	case 'N':
+		if (expander.vendor)
+			vendor_exclusive();
+		expander.vendor = V_NOKIA;
+		break;
+	case 'n':
+		if (expander.vendor)
+			vendor_exclusive();
+		expander.vendor = V_NOKIA_MD;
+		break;
+	case 't':
+		if (expander.generation)
+			exclusive();
+		expander.generation = T_ASSET;
+		break;
+	case 'T':
+		pipelining = 0;
+		break;
+	case 's':
+		expander.sequence = 1;
+		break;
+	case 'S':
+		expander.sources = optarg;
+		break;
+	case 'U':
+		if (expander.vendor)
+			exclusive();
+		expander.vendor = V_HUAWEI;
+		break;
+	case 'u':
+		if (expander.vendor)
+			exclusive();
+		expander.vendor = V_HUAWEI_XPL;
+		break;
+	case 'W':
+		expander.aswidth = atoi(optarg);
+		if (expander.aswidth < 0) {
+			sx_report(SX_FATAL,"Invalid as-width: %s\n", optarg);
+			exit(1);
+		}
+		widthSet = 1;
+		break;
+	case 'w':
+		expander.validate_asns = 1;
+		break;
+	case 'X':
+		if (expander.vendor)
+			vendor_exclusive();
+		expander.vendor = V_CISCO_XR;
+		break;
+	case 'v':
+		version();
+		break;
+	case 'z':
+		if (expander.generation)
+			exclusive();
+		expander.generation = T_ROUTE_FILTER_LIST;
+		break;
+	default:
+		usage(1);
 	}
 	}
 
@@ -478,6 +482,13 @@ main(int argc, char* argv[])
 				expander.aswidth = 8;
 				break;
 			}
+		} else if (expander.generation == T_ASLIST) {
+			int vendor = expander.vendor;
+			switch (vendor) {
+			case V_JUNIPER:
+				expander.aswidth = 8;
+				break;
+			}
 		}
 	}
 
@@ -663,14 +674,18 @@ main(int argc, char* argv[])
 		    "only with Juniper route-filters\n");
 	}
 
-	if ((expander.generation == T_ASPATH || expander.generation == T_OASPATH)
+	if ((expander.generation == T_ASPATH
+	    || expander.generation == T_OASPATH
+	    || expander.generation == T_ASLIST)
 	    && af != AF_INET && !expander.validate_asns) {
-		sx_report(SX_FATAL, "Sorry, -6 makes no sense with as-path (-f/-G) "
+		sx_report(SX_FATAL, "Sorry, -6 makes no sense with as-path (-f/-G) or as-list (-H) "
 		    "generation\n");
 	}
 
-	if (expander.validate_asns && expander.generation != T_ASPATH
-	    && expander.generation != T_OASPATH) {
+	if (expander.validate_asns
+	    && expander.generation != T_ASPATH
+	    && expander.generation != T_OASPATH
+	    && expander.generation != T_ASLIST) {
 		sx_report(SX_FATAL, "Sorry, -w makes sense only for as-path "
 		    "(-f/-G) generation\n");
 	}
@@ -733,7 +748,7 @@ main(int argc, char* argv[])
 
 	switch (expander.generation) {
 		case T_NONE:
-			sx_report(SX_FATAL,"Unreachable point... call snar\n");
+			sx_report(SX_FATAL,"Unreachable point");
 			exit(1);
 		case T_ASPATH:
 			bgpq4_print_aspath(stdout, &expander);
@@ -741,6 +756,9 @@ main(int argc, char* argv[])
 		case T_OASPATH:
 			bgpq4_print_oaspath(stdout, &expander);
 			break;
+		case T_ASLIST:
+			bgpq4_print_aslist(stdout, &expander);
+			break;
 		case T_ASSET:
 			bgpq4_print_asset(stdout, &expander);
 			break;
