@@ -274,6 +274,45 @@ bgpq4_print_juniper_oaspath(FILE *f, struct bgpq_expander *b)
 	fprintf(f, " }\n}\n");
 }
 
+static void
+bgpq4_print_juniper_aslist(FILE *f, struct bgpq_expander *b)
+{
+	int			 nc = 0, lineNo = 0;
+	struct asn_entry	*asne, find, *res;
+
+	fprintf(f,"policy-options {\nreplace:\n as-list-group %s {\n",
+	    b->name);
+
+	find.asn = b->asnumber;
+	if ((res = RB_FIND(asn_tree, &b->asnlist, &find)) != NULL) {
+		fprintf(f, "  as-list a0 members %u;\n", res->asn);
+		RB_REMOVE(asn_tree, &b->asnlist, res);
+		lineNo++;
+	}
+
+	RB_FOREACH(asne, asn_tree, &b->asnlist) {
+		if (!nc) {
+			fprintf(f, "  as-list a%u members [ %u",
+			    lineNo, asne->asn);
+		} else {
+			fprintf(f," %u", asne->asn);
+		}
+
+		nc++;
+
+		if (nc == b->aswidth) {
+			fprintf(f, " ];\n");
+			nc = 0;
+			lineNo++;
+		}
+	}
+
+	if (nc)
+		fprintf(f, " ];\n");
+
+	fprintf(f, " }\n}\n");
+}
+
 static void
 bgpq4_print_openbgpd_oaspath(FILE *f, struct bgpq_expander *b)
 {
@@ -406,6 +445,43 @@ bgpq4_print_huawei_aspath(FILE *f, struct bgpq_expander *b)
 		fprintf(f, ")$\n");
 }
 
+static void
+bgpq4_print_huawei_xpl_aspath(FILE *f, struct bgpq_expander *b)
+{
+	int 			 nc = 0, comma = 1;
+	struct asn_entry	*asne, find, *res;
+
+	fprintf(f, "xpl as-path-list %s", b->name);
+
+	find.asn = b->asnumber;
+	if ((res = RB_FIND(asn_tree, &b->asnlist, &find)) != NULL) {
+		fprintf(f, "\n  regular ^%u(_%u)*$", res->asn, res->asn);
+		RB_REMOVE(asn_tree, &b->asnlist, res);
+	}
+
+	RB_FOREACH(asne, asn_tree, &b->asnlist) {
+		if (!nc) {
+			fprintf(f, "%s\n  regular ^%u(_[0-9]+)*_(%u",
+			    comma ? "," : "",
+			    b->asnumber,
+			    asne->asn);
+			comma = 1;
+		} else
+			fprintf(f, "|%u", asne->asn);
+
+		nc++;
+		if (nc == b->aswidth) {
+			fprintf(f, ")$");
+			nc = 0;
+		}
+	}
+
+	if (nc)
+		fprintf(f, ")$");
+
+	fprintf(f, "\nend-list\n");
+}
+
 static void
 bgpq4_print_huawei_oaspath(FILE *f, struct bgpq_expander *b)
 {
@@ -445,6 +521,42 @@ bgpq4_print_huawei_oaspath(FILE *f, struct bgpq_expander *b)
 		fprintf(f, ")$\n");
 }
 
+static void
+bgpq4_print_huawei_xpl_oaspath(FILE *f, struct bgpq_expander *b)
+{
+	int 			 nc = 0, comma = 0;
+	struct asn_entry	*asne, find, *res;
+
+	fprintf(f, "xpl as-path-list %s", b->name);
+
+	find.asn = b->asnumber;
+	if ((res = RB_FIND(asn_tree, &b->asnlist, &find)) != NULL) {
+		fprintf(f, "\n  regular ^(_%u)*$", res->asn);
+		RB_REMOVE(asn_tree, &b->asnlist, res);
+		comma = 1;
+	}
+
+	RB_FOREACH(asne, asn_tree, &b->asnlist) {
+		if (!nc) {
+			fprintf(f,"%s\n  regular ^(_[0-9]+)*_(%u",
+			    comma ? "," : "", asne->asn);
+			comma = 1;
+		} else
+			fprintf(f,"|%u",asne->asn);
+
+		nc++;
+		if (nc == b->aswidth) {
+			fprintf(f,")$");
+			nc = 0;
+		}
+	}
+
+	if (nc)
+		fprintf(f,")$");
+
+	fprintf(f,"\nend-list\n");
+}
+
 static void
 bgpq4_print_nokia_oaspath(FILE *f, struct bgpq_expander *b)
 {
@@ -664,7 +776,7 @@ bgpq4_print_bird_aspath(FILE* f, struct bgpq_expander* b)
 			nc = 0;
 	}
 
-	fprintf(f, "];\n");
+	fprintf(f, "\n];\n");
 }
 
 static void
@@ -763,6 +875,9 @@ bgpq4_print_aspath(FILE *f, struct bgpq_expander *b)
 	case V_HUAWEI:
 		bgpq4_print_huawei_aspath(f, b);
 		break;
+	case V_HUAWEI_XPL:
+		bgpq4_print_huawei_xpl_aspath(f, b);
+		break;
 	default:
 		sx_report(SX_FATAL,"Unknown vendor %i\n", b->vendor);
 	}
@@ -794,6 +909,21 @@ bgpq4_print_oaspath(FILE *f, struct bgpq_expander *b)
 	case V_HUAWEI:
 		bgpq4_print_huawei_oaspath(f, b);
 		break;
+	case V_HUAWEI_XPL:
+		bgpq4_print_huawei_xpl_oaspath(f, b);
+		break;
+	default:
+		sx_report(SX_FATAL,"Unknown vendor %i\n", b->vendor);
+	}
+}
+
+void
+bgpq4_print_aslist(FILE *f, struct bgpq_expander *b)
+{
+	switch (b->vendor) {
+	case V_JUNIPER:
+		bgpq4_print_juniper_aslist(f, b);
+		break;
 	default:
 		sx_report(SX_FATAL,"Unknown vendor %i\n", b->vendor);
 	}
@@ -973,6 +1103,43 @@ checkSon:
 		bgpq4_print_hprefix(n->son, ff);
 }
 
+static void
+bgpq4_print_hprefixxpl(struct sx_radix_node* n, void* ff)
+{
+	char prefix[128];
+	FILE* f = (FILE*)ff;
+
+	if (!f)
+		f = stdout;
+
+	if (n->isGlue)
+		goto checkSon;
+
+	sx_prefix_snprintf_sep(n->prefix, prefix, sizeof(prefix), " ");
+
+	if (n->isAggregate) {
+		if (n->aggregateLow>n->prefix->masklen) {
+			fprintf(f,"%s %s ge %u le %u",
+			    needscomma ? ",\n " : " ",
+			    prefix, n->aggregateLow, n->aggregateHi);
+		} else {
+			fprintf(f,"%s %s le %u",
+			    needscomma ? ",\n " : " ",
+			    prefix, n->aggregateHi);
+		}
+	} else {
+		fprintf(f, "%s %s",
+		    needscomma ? ",\n " : " ",
+		    prefix);
+	}
+
+	needscomma = 1;
+
+checkSon:
+	if (n->son)
+		bgpq4_print_hprefixxpl(n->son, ff);
+}
+
 static void
 bgpq4_print_eprefix(struct sx_radix_node *n, void *ff)
 {
@@ -1070,11 +1237,11 @@ bgpq4_print_ceacl(struct sx_radix_node *n, void *ff)
 		wildmask.s_addr = htonl(wildmask.s_addr);
 
 		if (wildaddr.s_addr) {
-			fprintf(f, "permit ip %s ",
+			fprintf(f, " permit ip %s ",
 			    inet_ntoa(n->prefix->addr.addr));
 			fprintf(f, "%s ", inet_ntoa(wildaddr));
 		} else {
-			fprintf(f, "permit ip host %s ",
+			fprintf(f, " permit ip host %s ",
 			    inet_ntoa(n->prefix->addr.addr));
 		}
 
@@ -1085,7 +1252,7 @@ bgpq4_print_ceacl(struct sx_radix_node *n, void *ff)
 			fprintf(f, "host %s\n", inet_ntoa(mask));
 		}
 	} else {
-		fprintf(f, "permit ip host %s host %s\n", prefix,
+		fprintf(f, " permit ip host %s host %s\n", prefix,
 		    inet_ntoa(netmask));
 	}
 
@@ -1369,6 +1536,18 @@ bgpq4_print_huawei_prefixlist(FILE *f, struct bgpq_expander *b)
 	}
 }
 
+static void
+bgpq4_print_huawei_xpl_prefixlist(FILE* f, struct bgpq_expander* b)
+{
+	bname = b->name ? b->name : "NN";
+
+	fprintf(f, "no xpl %s-prefix-list %s\nxpl %s-prefix-list %s\n", b->family==AF_INET ? "ip" : "ipv6", bname, b->family==AF_INET ? "ip" : "ipv6", bname);
+
+	sx_radix_tree_foreach(b->tree, bgpq4_print_hprefixxpl, f);
+
+	fprintf(f, "\nend-list\n");
+}
+
 static void
 bgpq4_print_arista_prefixlist(FILE *f, struct bgpq_expander *b)
 {
@@ -1615,6 +1794,9 @@ bgpq4_print_prefixlist(FILE *f, struct bgpq_expander *b)
 	case V_HUAWEI:
 		bgpq4_print_huawei_prefixlist(f, b);
 		break;
+	case V_HUAWEI_XPL:
+		bgpq4_print_huawei_xpl_prefixlist(f, b);
+		break;
 	case V_MIKROTIK:
 		bgpq4_print_mikrotik_prefixlist(f, b);
 		break;
@@ -1632,6 +1814,7 @@ bgpq4_print_eacl(FILE *f, struct bgpq_expander *b)
 		bgpq4_print_juniper_routefilter(f, b);
 		break;
 	case V_CISCO:
+	case V_ARISTA:
 		bgpq4_print_cisco_eacl(f, b);
 		break;
 	case V_OPENBGPD:
@@ -1643,9 +1826,6 @@ bgpq4_print_eacl(FILE *f, struct bgpq_expander *b)
 	case V_NOKIA_MD:
 		bgpq4_print_nokia_md_prefixlist(f, b);
 		break;
-	case V_ARISTA:
-		bgpq4_print_cisco_eacl(f, b);
-		break;
 	default:
 		sx_report(SX_FATAL, "unreachable point\n");
 	}
